import functools
import inspect
from contextlib import asynccontextmanager
from contextvars import ContextVar
from inspect import Parameter
from typing import TYPE_CHECKING, Any, final

from fastapi.routing import APIRoute
from starlette.requests import Request
from starlette.routing import Match
from starlette.websockets import WebSocket

from aspy_dependency_injection._service_lookup._typed_type import TypedType
from aspy_dependency_injection.injectable import Injectable

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable, Sequence

    from fastapi import FastAPI
    from starlette.routing import BaseRoute
    from starlette.types import ASGIApp, Receive, Scope, Send

    from aspy_dependency_injection.service_collection import ServiceCollection
    from aspy_dependency_injection.service_provider import (
        ServiceProvider,
        ServiceScope,
    )

current_request: ContextVar[Request | WebSocket] = ContextVar("aspy_starlette_request")


@final
class FastApiDependencyInjection:
    @classmethod
    def setup(cls, app: FastAPI, services: ServiceCollection) -> None:
        service_provider = services.build_service_provider()
        app.state.aspy_service_provider = service_provider
        app.add_middleware(_AspyAsgiMiddleware)
        cls._update_lifespan(app, service_provider)
        cls._inject_routes(app.routes)

    @classmethod
    def _update_lifespan(cls, app: FastAPI, service_provider: ServiceProvider) -> None:
        old_lifespan = app.router.lifespan_context

        @asynccontextmanager
        async def new_lifespan(app: FastAPI) -> AsyncGenerator[Any]:
            async with old_lifespan(app) as state:
                yield state

            await service_provider.__aexit__(None, None, None)

        app.router.lifespan_context = new_lifespan

    @classmethod
    def _are_annotated_parameters_with_aspy_dependencies(
        cls, target: Callable[..., Any]
    ) -> bool:
        for parameter in inspect.signature(target).parameters.values():
            if parameter.annotation is not None and isinstance(
                parameter.annotation, Injectable
            ):
                return True

        return False

    @classmethod
    def _inject_routes(cls, routes: list[BaseRoute]) -> None:
        for route in routes:
            if not (
                isinstance(route, APIRoute)
                and route.dependant.call is not None
                and inspect.iscoroutinefunction(route.dependant.call)
                and not cls._are_annotated_parameters_with_aspy_dependencies(
                    route.dependant.call
                )
            ):
                continue

            route.dependant.call = cls._inject_from_container(route.dependant.call)

    @classmethod
    def _inject_from_container(cls, target: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(target)
        async def _inject_async_target(*args: Any, **kwargs: Any) -> Any:  # noqa: ANN401
            parameters_to_inject = cls._get_parameters_to_inject(target)
            parameters_to_inject_resolved: dict[str, Any] = {
                injected_parameter_name: await cls._get_request_container().service_provider.get_service_object(
                    TypedType.from_type(injected_parameter_class)
                )
                for injected_parameter_name, injected_parameter_class in parameters_to_inject.items()
            }
            return await target(*args, **{**kwargs, **parameters_to_inject_resolved})

        return _inject_async_target

    @classmethod
    def _get_request_container(cls) -> ServiceScope:
        """When inside a request, returns the scoped container instance handling the current request.

        This is what you almost always want.It has all the information the app container has in addition
        to data specific to the current request.
        """
        return current_request.get().state.aspy_service_scope

    @classmethod
    def _get_parameters_to_inject(cls, target: Callable[..., Any]) -> dict[str, type]:
        result: dict[str, type] = {}

        for parameter_name, parameter in inspect.signature(target).parameters.items():
            if parameter.annotation is Parameter.empty:
                continue

            injectable_dependency = cls._get_injectable_dependency(
                parameter.annotation.__metadata__
            )

            if injectable_dependency is None:
                continue

            service_type = parameter.annotation.__args__[0]
            result[parameter_name] = service_type

        return result

    @classmethod
    def _get_injectable_dependency(cls, metadata: Sequence[Any]) -> Injectable | None:
        for metadata_item in metadata:
            if hasattr(metadata_item, "dependency"):
                dependency = metadata_item.dependency()  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType, reportAttributeAccessIssue]

                if isinstance(dependency, Injectable):
                    return dependency

        return None


@final
class _AspyAsgiMiddleware:
    def __init__(self, app: ASGIApp) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in {"http", "websocket"}:
            return await self.app(scope, receive, send)

        if scope["type"] == "http":
            request = Request(scope, receive=receive, send=send)
        else:
            request = WebSocket(scope, receive, send)

        token = current_request.set(request)

        try:
            is_async_endpoint = False

            for route in scope["app"].routes:
                if (
                    isinstance(route, APIRoute)
                    and route.matches(scope)[0] == Match.FULL
                ):
                    original = inspect.unwrap(route.dependant.call)  # pyright: ignore[reportArgumentType]
                    is_async_endpoint = inspect.iscoroutinefunction(original)

                    if is_async_endpoint:
                        break

            if not is_async_endpoint:
                await self.app(scope, receive, send)
                return None

            service_provider: ServiceProvider = request.app.state.aspy_service_provider

            async with service_provider.create_scope() as service_scope:
                request.state.aspy_service_scope = service_scope
                await self.app(scope, receive, send)
        finally:
            current_request.reset(token)



from inspect import Parameter
from typing import Final, final

from aspy_dependency_injection._service_lookup._typed_type import TypedType


@final
class ParameterInformation:
    _parameter_type: Final[TypedType]

    def __init__(self, parameter: Parameter, type_: type) -> None:
        if parameter.annotation is Parameter.empty:
            error_message = f"The parameter '{parameter.name}' of the class '{type_}' must have a type annotation"
            raise RuntimeError(error_message)

        self._parameter_type = TypedType.from_type(parameter.annotation)

    @property
    def parameter_type(self) -> TypedType:
        return self._parameter_type



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, cast

from aspy_dependency_injection._service_lookup._call_site_kind import CallSiteKind
from aspy_dependency_injection._service_lookup.call_site_result_cache_location import (
    CallSiteResultCacheLocation,
)

if TYPE_CHECKING:
    from aspy_dependency_injection._service_lookup._async_factory_call_site import (
        AsyncFactoryCallSite,
    )
    from aspy_dependency_injection._service_lookup._constructor_call_site import (
        ConstructorCallSite,
    )
    from aspy_dependency_injection._service_lookup._service_call_site import (
        ServiceCallSite,
    )
    from aspy_dependency_injection._service_lookup._service_provider_call_site import (
        ServiceProviderCallSite,
    )
    from aspy_dependency_injection._service_lookup._sync_factory_call_site import (
        SyncFactoryCallSite,
    )


class CallSiteVisitor[TArgument, TResult](ABC):
    async def _visit_call_site(
        self, call_site: ServiceCallSite, argument: TArgument
    ) -> TResult:
        match call_site.cache.location:
            case CallSiteResultCacheLocation.ROOT:
                return await self._visit_root_cache(call_site, argument)
            case CallSiteResultCacheLocation.SCOPE:
                return await self._visit_scope_cache(call_site, argument)
            case CallSiteResultCacheLocation.DISPOSE:
                return await self._visit_dispose_cache(call_site, argument)
            case CallSiteResultCacheLocation.NONE:
                return await self._visit_no_cache(call_site, argument)

    async def _visit_root_cache(
        self, call_site: ServiceCallSite, argument: TArgument
    ) -> TResult:
        return await self._visit_call_site_main(call_site, argument)

    async def _visit_scope_cache(
        self, call_site: ServiceCallSite, argument: TArgument
    ) -> TResult:
        return await self._visit_call_site_main(call_site, argument)

    async def _visit_dispose_cache(
        self, call_site: ServiceCallSite, argument: TArgument
    ) -> TResult:
        return await self._visit_call_site_main(call_site, argument)

    async def _visit_no_cache(
        self, call_site: ServiceCallSite, argument: TArgument
    ) -> TResult:
        return await self._visit_call_site_main(call_site, argument)

    async def _visit_call_site_main(
        self, call_site: ServiceCallSite, argument: TArgument
    ) -> TResult:
        match call_site.kind:
            case CallSiteKind.SYNC_FACTORY:
                return await self._visit_sync_factory(
                    sync_factory_call_site=cast("SyncFactoryCallSite", call_site),
                    argument=argument,
                )
            case CallSiteKind.ASYNC_FACTORY:
                return await self._visit_async_factory(
                    async_factory_call_site=cast("AsyncFactoryCallSite", call_site),
                    argument=argument,
                )
            case CallSiteKind.CONSTRUCTOR:
                return await self._visit_constructor(
                    constructor_call_site=cast("ConstructorCallSite", call_site),
                    argument=argument,
                )
            case CallSiteKind.SERVICE_PROVIDER:
                return self._visit_service_provider(
                    service_provider_call_site=cast(
                        "ServiceProviderCallSite", call_site
                    ),
                    argument=argument,
                )

    @abstractmethod
    async def _visit_constructor(
        self, constructor_call_site: ConstructorCallSite, argument: TArgument
    ) -> TResult: ...

    @abstractmethod
    async def _visit_sync_factory(
        self, sync_factory_call_site: SyncFactoryCallSite, argument: TArgument
    ) -> TResult: ...

    @abstractmethod
    async def _visit_async_factory(
        self, async_factory_call_site: AsyncFactoryCallSite, argument: TArgument
    ) -> TResult: ...

    @abstractmethod
    def _visit_service_provider(
        self, service_provider_call_site: ServiceProviderCallSite, argument: TArgument
    ) -> TResult: ...



from abc import ABC, abstractmethod
from typing import cast

from aspy_dependency_injection._service_lookup._typed_type import TypedType


class BaseServiceProvider(ABC):
    """Define a mechanism for retrieving a service object; that is, an object that provides custom support to other objects."""

    @abstractmethod
    async def get_service_object(self, service_type: TypedType) -> object | None: ...

    async def get_service[TService](
        self, service_type: type[TService]
    ) -> TService | None:
        service = await self.get_service_object(TypedType.from_type(service_type))

        if service is None:
            return None

        return cast("TService", service)

    async def get_required_service[TService](
        self, service_type: type[TService]
    ) -> TService:
        service = await self.get_service_object(TypedType.from_type(service_type))
        assert service is not None
        return cast("TService", service)



from abc import ABC, abstractmethod
from contextlib import AbstractAsyncContextManager
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aspy_dependency_injection.abstractions.base_service_provider import (
        BaseServiceProvider,
    )


class ServiceScope(AbstractAsyncContextManager["ServiceScope"], ABC):
    """Defines a disposable service scope.

    The __aexit__ method ends the scope lifetime. Once called, any scoped
    services that have been resolved from ServiceProvider will be disposed.
    """

    @property
    @abstractmethod
    def service_provider(self) -> BaseServiceProvider:
        """Gets the :class:`BaseServiceProvider` used to resolve dependencies from the scope."""



from typing import TYPE_CHECKING, Final, final, override

from aspy_dependency_injection._service_lookup._call_site_kind import CallSiteKind
from aspy_dependency_injection._service_lookup._service_call_site import ServiceCallSite

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from aspy_dependency_injection._service_lookup._result_cache import ResultCache
    from aspy_dependency_injection._service_lookup._typed_type import TypedType


@final
class AsyncFactoryCallSite(ServiceCallSite):
    _service_type: Final[TypedType]
    _implementation_factory: Final[Callable[..., Awaitable[object]]]

    def __init__(
        self,
        cache: ResultCache,
        service_type: TypedType,
        implementation_factory: Callable[..., Awaitable[object]],
    ) -> None:
        super().__init__(cache)
        self._service_type = service_type
        self._implementation_factory = implementation_factory

    @property
    @override
    def service_type(self) -> TypedType:
        return self._service_type

    @property
    @override
    def kind(self) -> CallSiteKind:
        return CallSiteKind.ASYNC_FACTORY

    @property
    def implementation_factory(
        self,
    ) -> Callable[..., Awaitable[object]]:
        return self._implementation_factory



from typing import ClassVar, final


@final
class AspyUndefined:
    """A type used as a sentinel for undefined values."""

    INSTANCE: ClassVar[AspyUndefined]


AspyUndefined.INSTANCE = AspyUndefined()



import asyncio
from typing import TYPE_CHECKING, Final

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable


class AsyncConcurrentDictionary[TKey, TValue]:
    """Coroutine-safe collection of key/value pairs that can be accessed by multiple coroutines concurrently."""

    _dict: Final[dict[TKey, TValue]]
    _lock: Final[asyncio.Lock]

    def __init__(self) -> None:
        self._dict = {}
        self._lock = asyncio.Lock()

    async def get_or_add(
        self, key: TKey, value_factory: Callable[[TKey], Awaitable[TValue]]
    ) -> TValue:
        if key not in self._dict:
            async with self._lock:
                if key not in self._dict:
                    value = await value_factory(key)
                    self._dict[key] = value

        return self._dict[key]

    def get(self, key: TKey) -> TValue | None:
        return self._dict.get(key)

    def __getitem__(self, key: TKey, /) -> TValue:
        return self._dict[key]

    async def upsert(self, key: TKey, value: TValue) -> None:
        async with self._lock:
            self._dict[key] = value



import asyncio
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from aspy_dependency_injection._service_lookup._call_site_kind import CallSiteKind
    from aspy_dependency_injection._service_lookup._result_cache import ResultCache
    from aspy_dependency_injection._service_lookup._typed_type import TypedType


class ServiceCallSite(ABC):
    """Representation of how a service must be created."""

    _cache: ResultCache
    _value: object | None
    _lock: asyncio.Lock

    def __init__(self, cache: ResultCache) -> None:
        self._cache = cache
        self._value = None
        self._lock = asyncio.Lock()

    @property
    def cache(self) -> ResultCache:
        return self._cache

    @property
    def value(self) -> object | None:
        return None

    @value.setter
    def value(self, value: object | None) -> None:
        self._value = value

    @property
    def lock(self) -> asyncio.Lock:
        return self._lock

    @property
    @abstractmethod
    def service_type(self) -> TypedType: ...

    @property
    @abstractmethod
    def kind(self) -> CallSiteKind: ...



import asyncio
from typing import TYPE_CHECKING, ClassVar, Final, final

from aspy_dependency_injection._async_concurrent_dictionary import (
    AsyncConcurrentDictionary,
)
from aspy_dependency_injection._service_lookup._async_factory_call_site import (
    AsyncFactoryCallSite,
)
from aspy_dependency_injection._service_lookup._constructor_call_site import (
    ConstructorCallSite,
)
from aspy_dependency_injection._service_lookup._constructor_information import (
    ConstructorInformation,
)
from aspy_dependency_injection._service_lookup._result_cache import ResultCache
from aspy_dependency_injection._service_lookup._service_call_site import (
    ServiceCallSite,
)
from aspy_dependency_injection._service_lookup._service_identifier import (
    ServiceIdentifier,
)
from aspy_dependency_injection._service_lookup._sync_factory_call_site import (
    SyncFactoryCallSite,
)
from aspy_dependency_injection._service_lookup.service_cache_key import ServiceCacheKey

if TYPE_CHECKING:
    from aspy_dependency_injection._service_lookup._call_site_chain import CallSiteChain
    from aspy_dependency_injection._service_lookup._parameter_information import (
        ParameterInformation,
    )
    from aspy_dependency_injection._service_lookup._typed_type import TypedType
    from aspy_dependency_injection.service_collection import ServiceCollection
    from aspy_dependency_injection.service_descriptor import ServiceDescriptor


@final
class CallSiteFactory:
    _DEFAULT_SLOT: ClassVar[int] = 0

    _descriptors: Final[list[ServiceDescriptor]]
    _descriptor_lookup: Final[dict[ServiceIdentifier, _ServiceDescriptorCacheItem]]
    _call_site_cache: Final[AsyncConcurrentDictionary[ServiceCacheKey, ServiceCallSite]]
    _call_site_locks: Final[AsyncConcurrentDictionary[ServiceIdentifier, asyncio.Lock]]

    def __init__(self, services: ServiceCollection) -> None:
        self._descriptors = services.descriptors.copy()
        self._descriptor_lookup = {}
        self._call_site_cache = AsyncConcurrentDictionary[
            ServiceCacheKey, ServiceCallSite
        ]()
        self._call_site_locks = AsyncConcurrentDictionary[
            ServiceIdentifier, asyncio.Lock
        ]()
        self._populate()

    async def get_call_site(
        self, service_identifier: ServiceIdentifier, call_site_chain: CallSiteChain
    ) -> ServiceCallSite | None:
        service_cache_key = ServiceCacheKey(service_identifier, self._DEFAULT_SLOT)
        service_call_site = self._call_site_cache.get(service_cache_key)

        if service_call_site is None:
            return await self._create_call_site(
                service_identifier=service_identifier, call_site_chain=call_site_chain
            )

        return service_call_site

    async def add(
        self, service_identifier: ServiceIdentifier, service_call_site: ServiceCallSite
    ) -> None:
        cache_key = ServiceCacheKey(service_identifier, self._DEFAULT_SLOT)
        await self._call_site_cache.upsert(key=cache_key, value=service_call_site)

    async def _create_call_site(
        self, service_identifier: ServiceIdentifier, call_site_chain: CallSiteChain
    ) -> ServiceCallSite | None:
        async def _create_new_lock(_: ServiceIdentifier) -> asyncio.Lock:
            return asyncio.Lock()

        # We need to lock the resolution process for a single service type at a time.
        # Consider the following:
        # C -> D -> A
        # E -> D -> A
        # Resolving C and E in parallel means that they will be modifying the callsite cache concurrently
        # to add the entry for C and E, but the resolution of D and A is synchronized
        # to make sure C and E both reference the same instance of the callsite.
        #
        # This is to make sure we can safely store singleton values on the callsites themselves

        call_site_lock = await self._call_site_locks.get_or_add(
            service_identifier, _create_new_lock
        )

        # Check if the lock is already acquired to prevent deadlocks in case of re-entrancy
        if call_site_lock.locked():
            call_site_chain.check_circular_dependency(service_identifier)

        async with call_site_lock:
            return await self._try_create_exact_from_service_identifier(
                service_identifier, call_site_chain
            )

    def _populate(self) -> None:
        for descriptor in self._descriptors:
            cache_key = ServiceIdentifier.from_descriptor(descriptor)
            cache_item = self._descriptor_lookup.get(
                cache_key, _ServiceDescriptorCacheItem()
            )
            self._descriptor_lookup[cache_key] = cache_item.add(descriptor)

    async def _try_create_exact_from_service_identifier(
        self, service_identifier: ServiceIdentifier, call_site_chain: CallSiteChain
    ) -> ServiceCallSite | None:
        service_descriptor_cache_item = self._descriptor_lookup.get(
            service_identifier, None
        )

        if service_descriptor_cache_item is not None:
            return await self._try_create_exact_from_service_descriptor(
                service_descriptor_cache_item.last,
                service_identifier,
                call_site_chain,
                self._DEFAULT_SLOT,
            )

        return None

    async def _try_create_exact_from_service_descriptor(
        self,
        service_descriptor: ServiceDescriptor,
        service_identifier: ServiceIdentifier,
        call_site_chain: CallSiteChain,
        slot: int,
    ) -> ServiceCallSite | None:
        if not self._should_create_exact(
            service_descriptor.service_type, service_identifier.service_type
        ):
            return None

        return await self._create_exact(
            service_descriptor, service_identifier, call_site_chain, slot
        )

    def _should_create_exact(
        self, descriptor_type: TypedType, service_type: TypedType
    ) -> bool:
        return descriptor_type == service_type

    async def _create_exact(
        self,
        service_descriptor: ServiceDescriptor,
        service_identifier: ServiceIdentifier,
        call_site_chain: CallSiteChain,
        slot: int,
    ) -> ServiceCallSite:
        call_site_key = ServiceCacheKey(service_identifier, slot)
        service_call_site = self._call_site_cache.get(call_site_key)

        if service_call_site is not None:
            return service_call_site

        cache = ResultCache.from_lifetime(
            service_descriptor.lifetime, service_identifier, slot
        )

        if service_descriptor.sync_implementation_factory is not None:
            assert service_descriptor.sync_implementation_factory is not None
            service_call_site = SyncFactoryCallSite(
                cache=cache,
                service_type=service_identifier.service_type,
                implementation_factory=service_descriptor.sync_implementation_factory,
            )
        elif service_descriptor.async_implementation_factory is not None:
            assert service_descriptor.async_implementation_factory is not None
            service_call_site = AsyncFactoryCallSite(
                cache=cache,
                service_type=service_identifier.service_type,
                implementation_factory=service_descriptor.async_implementation_factory,
            )
        elif service_descriptor.has_implementation_type():
            assert service_descriptor.implementation_type is not None
            service_call_site = await self._create_constructor_call_site(
                cache=cache,
                service_identifier=service_identifier,
                implementation_type=service_descriptor.implementation_type,
                call_site_chain=call_site_chain,
            )
        else:
            error_message = "Invalid service descriptor"
            raise RuntimeError(error_message)

        await self._call_site_cache.upsert(key=call_site_key, value=service_call_site)
        return service_call_site

    async def _create_constructor_call_site(
        self,
        cache: ResultCache,
        service_identifier: ServiceIdentifier,
        implementation_type: TypedType,
        call_site_chain: CallSiteChain,
    ) -> ServiceCallSite:
        try:
            call_site_chain.add(service_identifier, implementation_type)
            parameter_call_sites: list[ServiceCallSite] | None = None
            constructor_information = ConstructorInformation(implementation_type)
            parameters = constructor_information.get_parameters()
            parameter_call_sites = await self._create_argument_call_sites(
                parameters, call_site_chain
            )
            return ConstructorCallSite(
                cache=cache,
                service_type=service_identifier.service_type,
                constructor_information=constructor_information,
                parameter_call_sites=parameter_call_sites,
            )
        finally:
            call_site_chain.remove(service_identifier)

    async def _create_argument_call_sites(
        self, parameters: list[ParameterInformation], call_site_chain: CallSiteChain
    ) -> list[ServiceCallSite]:
        if len(parameters) == 0:
            return []

        parameter_call_sites: list[ServiceCallSite] = []

        for parameter in parameters:
            call_site = await self._create_call_site(
                ServiceIdentifier.from_service_type(parameter.parameter_type),
                call_site_chain,
            )
            assert call_site is not None
            parameter_call_sites.append(call_site)

        return parameter_call_sites


@final
class _ServiceDescriptorCacheItem:
    _item: ServiceDescriptor | None
    _items: list[ServiceDescriptor] | None

    def __init__(self) -> None:
        self._item = None
        self._items = None

    @property
    def last(self) -> ServiceDescriptor:
        if self._items is not None and len(self._items) > 0:
            return self._items[len(self._items) - 1]

        assert self._item is not None
        return self._item

    def add(self, descriptor: ServiceDescriptor) -> _ServiceDescriptorCacheItem:
        new_cache_item = _ServiceDescriptorCacheItem()

        if self._item is None:
            new_cache_item._item = descriptor
        else:
            new_cache_item._item = self._item
            new_cache_item._items = self._items if self._items is not None else []
            new_cache_item._items.append(descriptor)

        return new_cache_item



from collections.abc import Hashable
from typing import (
    TYPE_CHECKING,
    Final,
    final,
    override,
)

if TYPE_CHECKING:
    from aspy_dependency_injection._service_lookup._typed_type import TypedType
    from aspy_dependency_injection.service_descriptor import ServiceDescriptor


@final
class ServiceIdentifier(Hashable):
    """Internal registered service during resolution."""

    _service_type: Final[TypedType]

    def __init__(self, service_type: TypedType) -> None:
        self._service_type = service_type

    @property
    def service_type(self) -> TypedType:
        return self._service_type

    @classmethod
    def from_service_type(cls, service_type: TypedType) -> ServiceIdentifier:
        return cls(service_type)

    @classmethod
    def from_descriptor(
        cls, service_descriptor: ServiceDescriptor
    ) -> ServiceIdentifier:
        return cls(service_descriptor.service_type)

    @override
    def __hash__(self) -> int:
        return hash(self._service_type)

    def __eq__(self, value: object) -> bool:
        if not isinstance(value, ServiceIdentifier):
            return NotImplemented

        return self._service_type == value._service_type



import asyncio
from dataclasses import dataclass
from typing import TYPE_CHECKING, Final, Self, final, override

from aspy_dependency_injection._async_concurrent_dictionary import (
    AsyncConcurrentDictionary,
)
from aspy_dependency_injection._service_lookup._call_site_chain import CallSiteChain
from aspy_dependency_injection._service_lookup._call_site_factory import CallSiteFactory
from aspy_dependency_injection._service_lookup._runtime_service_provider_engine import (
    RuntimeServiceProviderEngine,
)
from aspy_dependency_injection._service_lookup._service_identifier import (
    ServiceIdentifier,
)
from aspy_dependency_injection._service_lookup._service_provider_call_site import (
    ServiceProviderCallSite,
)
from aspy_dependency_injection._service_lookup._typed_type import TypedType
from aspy_dependency_injection.abstractions.base_service_provider import (
    BaseServiceProvider,
)
from aspy_dependency_injection.abstractions.service_scope import (
    AbstractAsyncContextManager,
)
from aspy_dependency_injection.exceptions import ObjectDisposedError
from aspy_dependency_injection.service_provider_engine_scope import (
    ServiceProviderEngineScope,
)

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable
    from types import TracebackType

    from aspy_dependency_injection._service_lookup._service_call_site import (
        ServiceCallSite,
    )
    from aspy_dependency_injection._service_lookup._service_provider_engine import (
        ServiceProviderEngine,
    )
    from aspy_dependency_injection.abstractions.service_scope import ServiceScope
    from aspy_dependency_injection.service_collection import ServiceCollection


@dataclass(frozen=True)
class _ServiceAccessor:
    call_site: ServiceCallSite | None
    realized_service: Callable[[ServiceProviderEngineScope], Awaitable[object | None]]


@final
class ServiceProvider(
    BaseServiceProvider, AbstractAsyncContextManager["ServiceProvider"]
):
    """Provider that resolves services."""

    _services: Final[ServiceCollection]
    _root: Final[ServiceProviderEngineScope]
    _engine: Final[ServiceProviderEngine]
    _service_accessors: Final[
        AsyncConcurrentDictionary[ServiceIdentifier, _ServiceAccessor]
    ]
    _is_disposed: bool
    _call_site_factory: Final[CallSiteFactory]

    def __init__(self, services: ServiceCollection) -> None:
        self._services = services
        self._root = ServiceProviderEngineScope(
            service_provider=self, is_root_scope=True
        )
        self._engine = self._get_engine()
        self._service_accessors = AsyncConcurrentDictionary()
        self._is_disposed = False
        self._call_site_factory = CallSiteFactory(services)

    @property
    def root(self) -> ServiceProviderEngineScope:
        return self._root

    @property
    def is_disposed(self) -> bool:
        return self._is_disposed

    @override
    async def get_service_object(self, service_type: TypedType) -> object | None:
        if self._is_disposed:
            raise ObjectDisposedError

        return await self.get_service_from_service_identifier(
            service_identifier=ServiceIdentifier.from_service_type(service_type),
            service_provider_engine_scope=self._root,
        )

    def create_scope(self) -> ServiceScope:
        """Create a new :class:`ServiceScope` that can be used to resolve scoped services."""
        if self._is_disposed:
            raise ObjectDisposedError

        return ServiceProviderEngineScope(service_provider=self, is_root_scope=False)

    async def get_service_from_service_identifier(
        self,
        service_identifier: ServiceIdentifier,
        service_provider_engine_scope: ServiceProviderEngineScope,
    ) -> object | None:
        service_accessor = await self._service_accessors.get_or_add(
            key=service_identifier, value_factory=self._create_service_accessor
        )
        return await service_accessor.realized_service(service_provider_engine_scope)

    async def _create_service_accessor(
        self, service_identifier: ServiceIdentifier
    ) -> _ServiceAccessor:
        def realized_service_returning_none(
            _: ServiceProviderEngineScope,
        ) -> Awaitable[object | None]:
            future = asyncio.Future[None]()
            future.set_result(None)
            return future

        call_site = await self._call_site_factory.get_call_site(
            service_identifier, CallSiteChain()
        )

        if call_site is not None:
            realized_service = self._engine.realize_service(call_site)
            return _ServiceAccessor(
                call_site=call_site, realized_service=realized_service
            )

        return _ServiceAccessor(
            call_site=call_site, realized_service=realized_service_returning_none
        )

    def _get_engine(self) -> ServiceProviderEngine:
        return RuntimeServiceProviderEngine.INSTANCE

    @override
    async def __aenter__(self) -> Self:
        # Add built-in services that aren't part of the list of service descriptors
        await self._call_site_factory.add(
            ServiceIdentifier.from_service_type(
                TypedType.from_type(BaseServiceProvider)
            ),
            ServiceProviderCallSite(),
        )
        return self

    @override
    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> bool | None:
        self._is_disposed = True
        await self._root.__aexit__(exc_type, exc_val, exc_tb)



from dataclasses import dataclass
from enum import Flag
from typing import (
    TYPE_CHECKING,
    ClassVar,
    final,
    get_type_hints,
    override,
)

from aspy_dependency_injection._aspy_undefined import AspyUndefined
from aspy_dependency_injection._service_lookup._call_site_visitor import CallSiteVisitor
from aspy_dependency_injection._service_lookup._supports_async_context_manager import (
    SupportsAsyncContextManager,
)
from aspy_dependency_injection._service_lookup._supports_context_manager import (
    SupportsContextManager,
)
from aspy_dependency_injection._service_lookup._typed_type import TypedType

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from aspy_dependency_injection._service_lookup._async_factory_call_site import (
        AsyncFactoryCallSite,
    )
    from aspy_dependency_injection._service_lookup._constructor_call_site import (
        ConstructorCallSite,
    )
    from aspy_dependency_injection._service_lookup._service_call_site import (
        ServiceCallSite,
    )
    from aspy_dependency_injection._service_lookup._service_provider_call_site import (
        ServiceProviderCallSite,
    )
    from aspy_dependency_injection._service_lookup._sync_factory_call_site import (
        SyncFactoryCallSite,
    )
    from aspy_dependency_injection.service_provider_engine_scope import (
        ServiceProviderEngineScope,
    )


class _RuntimeResolverLock(Flag):
    NONE = 0
    SCOPE = 1
    ROOT = 2


@dataclass(frozen=True)
class RuntimeResolverContext:
    scope: ServiceProviderEngineScope
    acquired_locks: _RuntimeResolverLock


@final
class CallSiteRuntimeResolver(CallSiteVisitor[RuntimeResolverContext, object | None]):
    INSTANCE: ClassVar[CallSiteRuntimeResolver]

    async def resolve(
        self, call_site: ServiceCallSite, scope: ServiceProviderEngineScope
    ) -> object | None:
        return await self._visit_call_site(
            call_site,
            RuntimeResolverContext(
                scope=scope, acquired_locks=_RuntimeResolverLock.NONE
            ),
        )

    @override
    async def _visit_root_cache(
        self, call_site: ServiceCallSite, argument: RuntimeResolverContext
    ) -> object | None:
        # If the value is already calculated, return it directly
        if call_site.value is not None:
            return call_site.value

        lock_type = _RuntimeResolverLock.ROOT
        service_provider_engine_scope = argument.scope.root_provider.root

        async with call_site.lock:
            # Lock the callsite and check if another coroutine already cached the value
            if call_site.value is not None:
                return call_site.value

            resolved_service = await self._visit_call_site_main(
                call_site=call_site,
                argument=RuntimeResolverContext(
                    scope=service_provider_engine_scope,
                    acquired_locks=argument.acquired_locks | lock_type,
                ),
            )
            await service_provider_engine_scope.capture_disposable(resolved_service)
            call_site.value = resolved_service
            return resolved_service

    @override
    async def _visit_scope_cache(
        self, call_site: ServiceCallSite, argument: RuntimeResolverContext
    ) -> object | None:
        # Check if we are in the situation where scoped service was promoted to singleton
        # and we need to lock the root
        if argument.scope.is_root_scope:
            return await self._visit_root_cache(call_site, argument)

        return await self._visit_cache(
            call_site, argument, argument.scope, _RuntimeResolverLock.SCOPE
        )

    async def _visit_cache(
        self,
        call_site: ServiceCallSite,
        argument: RuntimeResolverContext,
        service_provider_engine_scope: ServiceProviderEngineScope,
        lock_type: _RuntimeResolverLock,
    ) -> object | None:
        is_lock_taken = False
        resolved_services_lock = service_provider_engine_scope.resolved_services_lock
        resolved_services = service_provider_engine_scope.realized_services

        # Taking locks only once allows us to fork resolution process
        # on another coroutine without causing the deadlock because we
        # always know that we are going to wait the other coroutine to finish before
        # releasing the lock
        if (argument.acquired_locks & lock_type) == _RuntimeResolverLock.NONE:
            await resolved_services_lock.acquire()
            is_lock_taken = True

        try:
            # Note: This method has already taken lock by the caller for resolution and access synchronization.
            # For scoped: takes a dictionary as both a resolution lock and a dictionary access lock.
            resolved_service = resolved_services.get(
                call_site.cache.key, AspyUndefined.INSTANCE
            )

            if resolved_service is not AspyUndefined.INSTANCE:
                return resolved_service

            resolved_service = await self._visit_call_site_main(
                call_site=call_site,
                argument=RuntimeResolverContext(
                    scope=service_provider_engine_scope,
                    acquired_locks=argument.acquired_locks | lock_type,
                ),
            )
            await service_provider_engine_scope.capture_disposable(resolved_service)
            resolved_services[call_site.cache.key] = resolved_service
            return resolved_service
        finally:
            if is_lock_taken:
                resolved_services_lock.release()

    @override
    async def _visit_dispose_cache(
        self, call_site: ServiceCallSite, argument: RuntimeResolverContext
    ) -> object | None:
        service = await self._visit_call_site_main(call_site, argument)
        return await argument.scope.capture_disposable(service)

    @override
    async def _visit_constructor(
        self,
        constructor_call_site: ConstructorCallSite,
        argument: RuntimeResolverContext,
    ) -> object:
        parameter_values: list[object | None] = [
            await self._visit_call_site(parameter_call_site, argument)
            for parameter_call_site in constructor_call_site.parameter_call_sites
        ]
        service = constructor_call_site.constructor_information.invoke(parameter_values)

        if isinstance(service, SupportsAsyncContextManager):
            await service.__aenter__()
        elif isinstance(service, SupportsContextManager):
            service.__enter__()

        return service

    @override
    async def _visit_sync_factory(
        self,
        sync_factory_call_site: SyncFactoryCallSite,
        argument: RuntimeResolverContext,
    ) -> object | None:
        parameter_services = await self.get_parameter_services(
            sync_factory_call_site.implementation_factory, argument.scope
        )
        service = sync_factory_call_site.implementation_factory(*parameter_services)

        if isinstance(service, SupportsAsyncContextManager):
            await service.__aenter__()
        elif isinstance(service, SupportsContextManager):
            service.__enter__()

        return service

    @override
    async def _visit_async_factory(
        self,
        async_factory_call_site: AsyncFactoryCallSite,
        argument: RuntimeResolverContext,
    ) -> object | None:
        parameter_services = await self.get_parameter_services(
            async_factory_call_site.implementation_factory, argument.scope
        )
        service = await async_factory_call_site.implementation_factory(
            *parameter_services
        )

        if isinstance(service, SupportsAsyncContextManager):
            await service.__aenter__()
        elif isinstance(service, SupportsContextManager):
            service.__enter__()

        return service

    @override
    def _visit_service_provider(
        self,
        service_provider_call_site: ServiceProviderCallSite,
        argument: RuntimeResolverContext,
    ) -> object | None:
        return argument.scope

    async def get_parameter_services(
        self,
        implementation_factory: Callable[..., Awaitable[object]]
        | Callable[..., object],
        scope: ServiceProviderEngineScope,
    ) -> list[object | None]:
        parameter_types = self._get_parameter_types(implementation_factory)
        parameter_services: list[object] = []

        for parameter_type in parameter_types:
            parameter_service = await scope.get_service_object(
                TypedType.from_type(parameter_type)
            )

            if parameter_service is None:
                error_message = (
                    f"Unable to resolve service for type '{parameter_type}' "
                    f"while attempting to invoke implementation factory "
                    f"'{implementation_factory}'."
                )
                raise RuntimeError(error_message)

            parameter_services.append(parameter_service)

        return parameter_services

    def _get_parameter_types(
        self,
        implementation_factory: Callable[..., Awaitable[object]]
        | Callable[..., object],
    ) -> list[type]:
        type_hints: dict[str, type] = get_type_hints(implementation_factory)
        return [
            parameter_type
            for parameter_name, parameter_type in type_hints.items()
            if parameter_name != "return"
        ]


CallSiteRuntimeResolver.INSTANCE = CallSiteRuntimeResolver()



import inspect
from typing import TYPE_CHECKING, Final, get_type_hints, overload

from aspy_dependency_injection._integrations._fastapi_dependency_injection import (
    FastApiDependencyInjection,
)
from aspy_dependency_injection.service_descriptor import ServiceDescriptor
from aspy_dependency_injection.service_lifetime import ServiceLifetime
from aspy_dependency_injection.service_provider import (
    ServiceProvider,
)

if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable

    from fastapi import FastAPI


class ServiceCollection:
    """Collection of service descriptors provided during configuration."""

    _descriptors: Final[list[ServiceDescriptor]]

    def __init__(self) -> None:
        self._descriptors = []

    @property
    def descriptors(self) -> list[ServiceDescriptor]:
        return self._descriptors

    @overload
    def add_transient[TService](self, service_type: type[TService], /) -> None: ...

    @overload
    def add_transient[TService](
        self,
        service_type: type[TService],
        implementation_factory: Callable[..., Awaitable[TService]],
        /,
    ) -> None: ...

    @overload
    def add_transient[TService](
        self,
        service_type: type[TService],
        implementation_factory: Callable[..., TService],
        /,
    ) -> None: ...

    @overload
    def add_transient[TService](
        self,
        implementation_factory: Callable[..., Awaitable[TService]],
        /,
    ) -> None: ...

    @overload
    def add_transient[TService](
        self,
        implementation_factory: Callable[..., TService],
        /,
    ) -> None: ...

    def add_transient[TService](
        self,
        service_type_or_implementation_factory: type[TService]
        | Callable[..., Awaitable[TService]]
        | Callable[..., TService],
        implementation_factory: Callable[..., Awaitable[TService]]
        | Callable[..., TService]
        | None = None,
        /,
    ) -> None:
        self._add_from_overloaded_constructor(
            ServiceLifetime.TRANSIENT,
            service_type_or_implementation_factory,
            implementation_factory,
        )

    @overload
    def add_singleton[TService](self, service_type: type[TService], /) -> None: ...

    @overload
    def add_singleton[TService](
        self,
        service_type: type[TService],
        implementation_factory: Callable[..., Awaitable[TService]],
        /,
    ) -> None: ...

    @overload
    def add_singleton[TService](
        self,
        service_type: type[TService],
        implementation_factory: Callable[..., TService],
        /,
    ) -> None: ...

    @overload
    def add_singleton[TService](
        self,
        implementation_factory: Callable[..., Awaitable[TService]],
        /,
    ) -> None: ...

    @overload
    def add_singleton[TService](
        self,
        implementation_factory: Callable[..., TService],
        /,
    ) -> None: ...

    def add_singleton[TService](
        self,
        service_type_or_implementation_factory: type[TService]
        | Callable[..., Awaitable[TService]]
        | Callable[..., TService],
        implementation_factory: Callable[..., Awaitable[TService]]
        | Callable[..., TService]
        | None = None,
        /,
    ) -> None:
        self._add_from_overloaded_constructor(
            ServiceLifetime.SINGLETON,
            service_type_or_implementation_factory,
            implementation_factory,
        )

    @overload
    def add_scoped[TService](self, service_type: type[TService], /) -> None: ...

    @overload
    def add_scoped[TService](
        self,
        service_type: type[TService],
        implementation_factory: Callable[..., Awaitable[TService]],
        /,
    ) -> None: ...

    @overload
    def add_scoped[TService](
        self,
        service_type: type[TService],
        implementation_factory: Callable[..., TService],
        /,
    ) -> None: ...

    @overload
    def add_scoped[TService](
        self,
        implementation_factory: Callable[..., Awaitable[TService]],
        /,
    ) -> None: ...

    @overload
    def add_scoped[TService](
        self,
        implementation_factory: Callable[..., TService],
        /,
    ) -> None: ...

    def add_scoped[TService](
        self,
        service_type_or_implementation_factory: type[TService]
        | Callable[..., Awaitable[TService]]
        | Callable[..., TService],
        implementation_factory: Callable[..., Awaitable[TService]]
        | Callable[..., TService]
        | None = None,
        /,
    ) -> None:
        self._add_from_overloaded_constructor(
            ServiceLifetime.SCOPED,
            service_type_or_implementation_factory,
            implementation_factory,
        )

    def build_service_provider(self) -> ServiceProvider:
        """Create a :class:`ServiceProvider` containing services from the provided :class:`ServiceCollection`."""
        return ServiceProvider(self)

    def configure_fastapi(self, app: FastAPI) -> None:
        """Configure the FastAPI application to use dependency injection using the services from this service collection."""
        FastApiDependencyInjection.setup(app, self)

    def _add_from_overloaded_constructor[TService](
        self,
        lifetime: ServiceLifetime,
        service_type_or_implementation_factory: type[TService]
        | Callable[..., Awaitable[TService]]
        | Callable[..., TService],
        implementation_factory: Callable[..., Awaitable[TService]]
        | Callable[..., TService]
        | None = None,
    ) -> None:
        service_type_to_add: type[TService] | None = None
        implementation_factory_to_add: (
            Callable[..., Awaitable[TService]] | Callable[..., TService] | None
        ) = None

        if isinstance(service_type_or_implementation_factory, type):
            service_type_to_add = service_type_or_implementation_factory

        if service_type_to_add is not None and implementation_factory is not None:
            implementation_factory_to_add = implementation_factory
        elif service_type_to_add is None and implementation_factory is None:
            implementation_factory_to_add = service_type_or_implementation_factory

        self._add(
            lifetime,
            service_type=service_type_to_add,
            implementation_factory=implementation_factory_to_add,
        )

    def _add[TService](
        self,
        lifetime: ServiceLifetime,
        service_type: type[TService] | None = None,
        implementation_factory: Callable[..., Awaitable[TService]]
        | Callable[..., TService]
        | None = None,
    ) -> None:
        provided_service_type = self._get_provided_service_type(
            service_type, implementation_factory
        )

        if implementation_factory is None:
            self._add_from_implementation_type(
                service_type=provided_service_type,
                implementation_type=provided_service_type,
                lifetime=lifetime,
            )
        elif inspect.iscoroutinefunction(implementation_factory):
            self._add_from_async_implementation_factory(
                service_type=provided_service_type,
                implementation_factory=implementation_factory,
                lifetime=lifetime,
            )
        else:
            self._add_from_sync_implementation_factory(
                service_type=provided_service_type,
                implementation_factory=implementation_factory,
                lifetime=lifetime,
            )

    def _get_provided_service_type[TService](
        self,
        service_type: type[TService] | None = None,
        implementation_factory: Callable[..., Awaitable[TService]]
        | Callable[..., TService]
        | None = None,
    ) -> type:
        if service_type is not None:
            return service_type

        assert implementation_factory is not None

        type_hints: dict[str, type] = get_type_hints(implementation_factory)
        return_type = type_hints.get("return")

        if return_type is None:
            error_message = "Missing return type hints from 'implementation_factory'"
            raise ValueError(error_message)

        return return_type

    def _add_from_implementation_type(
        self, service_type: type, implementation_type: type, lifetime: ServiceLifetime
    ) -> None:
        descriptor = ServiceDescriptor.from_implementation_type(
            service_type=service_type,
            implementation_type=implementation_type,
            lifetime=lifetime,
        )
        self._descriptors.append(descriptor)

    def _add_from_sync_implementation_factory(
        self,
        service_type: type,
        implementation_factory: Callable[..., object],
        lifetime: ServiceLifetime,
    ) -> None:
        descriptor = ServiceDescriptor.from_sync_implementation_factory(
            service_type=service_type,
            implementation_factory=implementation_factory,
            lifetime=lifetime,
        )
        self._descriptors.append(descriptor)

    def _add_from_async_implementation_factory(
        self,
        service_type: type,
        implementation_factory: Callable[..., Awaitable[object]],
        lifetime: ServiceLifetime,
    ) -> None:
        descriptor = ServiceDescriptor.from_async_implementation_factory(
            service_type=service_type,
            implementation_factory=implementation_factory,
            lifetime=lifetime,
        )
        self._descriptors.append(descriptor)



import inspect
from typing import TYPE_CHECKING, Final, final

from aspy_dependency_injection._service_lookup._parameter_information import (
    ParameterInformation,
)

if TYPE_CHECKING:
    from aspy_dependency_injection._service_lookup._typed_type import TypedType


@final
class ConstructorInformation:
    _type_: Final[TypedType]

    def __init__(self, type_: TypedType) -> None:
        self._type_ = type_

    def invoke(self, parameter_values: list[object]) -> object:
        return self._type_.invoke(parameter_values)

    def get_parameters(self) -> list[ParameterInformation]:
        init_method = self._type_.to_type().__init__
        init_signature = inspect.signature(init_method)
        return [
            ParameterInformation(parameter=parameter, type_=self._type_.to_type())
            for name, parameter in init_signature.parameters.items()
            if name not in ["self", "args", "kwargs"]
        ]
